//
//	Encoder for G.729A CS-ACELP Speech Codec
// 
//		dec_codebook_search.cpp : Module encapsulation for 
//			- Decoding Adaptive and Fixed Codebook
//

#include "StdAfx.h"
#include "dec_codebook_search.h"
#include "dec_adaptive_codebook_search.h"
#include "dec_fixed_codebook_search.h"

/*****************************************************************************
/*
/*	  Function to do memory update [A.3.10]
/*    Input: v[40] adaptive codebook vector 
/*           c[40] impulse response
/*    Output: u[200] filtered adaptive codebook vector
/*
/****************************************************************************/
void memory_update(float u[200], float g_p, float v[40], float g_c, float c[40])
{
	int n, i;
	int zero_index = 160;

	for (n=0; n<40; n++)		
		u[zero_index+n] = g_p*v[n] + g_c*c[n];
}

/*****************************************************************************
/*
/*	  Function for the main block of the adaptive codebook search
/*    Input: x[80]: target signal
/*			 h[80]: impulse response
/*			 u[240]: excitation buffer
/*			 r[40]: LP residue
/*			 T_op: Optimal open loop delay
/*    Output: *P0: Parity bit fr pitch delay
/*			  *P1: Pitch delay for the 1st subframe
/*			  *P2: Pitch delay for the 2nd subframe
/*			  *g_p1: adaptive codebook gain for the 1st subframe
/*		      *g_p2: adaptive codebook gain for the 2nd subframe
/*			  v[80]: adaptive codebook vector
/*			  y[80]: filtered adaptive codebook vector
/*
/****************************************************************************/
void dec_codebook_search( float u[240], int P0, int P1, int P2, int S1, int S2, int C1, int C2, float g_p1, float g_c1, float g_p2, float g_c2, float *beta, int *T1, int *T2, float c[80], float v[80])
{	
	int tmax, tmin;
	int frac1, frac2;	
	int n;
//	float v[80], c[80];

	if (P1 < 197)
	{
		*T1 = (P1 + 2)/3 + 19;
		frac1 = P1 - 3* *T1 + 58;
	}
	else
	{
		*T1 = P1 - 112;
		frac1 = 0;
	}

	tmin = *T1 - 5;
	if (tmin < 20)
		tmin = 20;
	tmax = tmin + 9;
	if (tmax > 143)
	{
		tmax = 143;
		tmin = tmax - 9;
	}
	*T2 = (P2+2)/3 -1 + tmin;
	frac2 = P2 - 2 - 3*((P2+2)/3-1);

	adaptive_codebook_vector(&u[0], *T1, frac1, &v[0]);
	dec_fixed_codebook(S1, C1, *T1, *beta, &c[0]);
	*beta = g_p1;
	memory_update(&u[0], g_p1, &v[0], g_c1, &c[0]);

	adaptive_codebook_vector(&u[40], *T2, frac2, &v[40]);
	dec_fixed_codebook(S2, C2, *T2, *beta, &c[40]);
	*beta = g_p2;
	memory_update(&u[40], g_p2, &v[40], g_c2, &c[40]);

	for (n=0; n<160; n++)
		u[n] = u[n+80];
}