//
//	Encoder for G.729A CS-ACELP Speech Codec
// 
//		encoder.cpp : Main program for the encoder 
//

#include "StdAfx.h"
#include "encoder.h"
#include "file_io.h"
#include "initialization.h"
#include "LP_analysis.h"
#include "pre_processing.h"
#include "open_loop.h"
#include "codebook_search.h"

int main(int argc, char* argv[])
{

	short *x_in, *y_out;

	int n;
	int freq = 0;
	int read;

	short *In, *Out;
	//unsigned char *Header;
	
	FILE *Infile=NULL, *Outfile=NULL;

	int size = 0;

	char fname[20];
	float s[240], s_p[240];
	float a[10], a_h[2][10];
	float q[2][10];
	float q_sf[2][10];
	float w[10];
	int i;
	int start;
	int current;
	int frame;

	int head=0;
	float r[80], h[80], x[80];
	float s_w[320];
	int T_op;
	bool P0;
	int P1, P2;
	float g_p1, g_p2;
	float v[80], u[240], y[80], c[80];
	int S1, S2, C1, C2;
	float g_c1 = 5.5847, g_c2;
	float U_h[4] = {-14.0, -14.0, -14.0, -14.0};

	FILE *fout2; 

	char foname[NUM_FILE][FILE_NAME_SIZE];
	strcpy(foname[0],"original.dat");
	strcpy(foname[1],"pre_processed_before_lowpass.dat");
	strcpy(foname[2],"pre_processed_after_lowpass.dat");
	strcpy(foname[3],"autocorrelate_func_before_BWE.dat");
	strcpy(foname[4],"autocorrelate_func_after_BWE.dat");
	strcpy(foname[5],"LP_synthesis.dat");
	strcpy(foname[6],"LSF.dat");

	FILE *fout[NUM_FILE];

	fout2 = fopen("coded_parameter.dat","w");

	for (i=0; i<=6; i++)
		fout[i] = fopen(foname[i],"w");

	if (argc != 2)
	{
		printf("\nA program to encode a 8kHz, 16-bit PCM wave file using CS-ACELP");
		printf("\nUsage: encoder [input wave file]");
	}
	else
	{
		strcpy(fname, argv[1]);

		//Opening input and output file
		if ((Infile = fopen(fname, "rb")) == NULL)
		{
			printf("Error opening the input wave file\n");		
			exit(-1);
		}
		
		else if ((Outfile = fopen("output.out", "wb")) == NULL)
		{
			printf("Error opening file for writing\n");
			fclose(Infile);
			exit(-1);
		}
		fclose(Outfile);
	
		checkHeader(Infile);

		//Allocate and initialize the memory
		x_in = (short *)malloc(BUFFERSIZE);
		y_out = (short *)malloc(BUFFERSIZE);
		for (n=0;n<240;n++)
		{
			x_in[n] = 0;
			y_out[n] = 0;
		}

		In = x_in;
		Out = y_out;
		n = 0;

		current = 0;
		initialize(q[current]);

		//Read in the initial frame
		fread(In, sizeof(short), SUBFRAMESIZE, Infile);
		preprocessing(y_out, x_in, n, SUBFRAMESIZE);
		frame = 0;		
		for (i=0; i<40; i++)
			fprintf(fout[0],"%f\n",(float)x_in[(n+i)%BUFFERSIZE]);		
		
		for (i=3*SUBFRAMESIZE; i<6*SUBFRAMESIZE; i++)
			s[i] = 0;
		start = 4*SUBFRAMESIZE;
		for (i=0; i<SUBFRAMESIZE; i++)
			s[i] = (float)y_out[i];

		n   = (n+SUBFRAMESIZE)%BUFFERSIZE;
		Outfile = fopen("output.out", "ab");
		fwrite(Out, sizeof(short), SUBFRAMESIZE, Outfile);
		fclose(Outfile);
		In  = x_in + n;
		Out = y_out + n;

		while((read = fread(In, sizeof(short), FRAMESIZE, Infile))==FRAMESIZE)
		{
			preprocessing(y_out, x_in, n, FRAMESIZE);
			frame++;			
			for (i=0; i<80; i++)
				fprintf(fout[0],"%f\n",(float)x_in[(n+i)%BUFFERSIZE]);
			
			start = (start+FRAMESIZE)%240;
			for (i=0; i<FRAMESIZE; i++)
				s[(start+SUBFRAMESIZE+i)%240] = (float)y_out[(n+i)%BUFFERSIZE];
			window_and_autocorrelate(s,start,s_p,a,fout);
			current = 1 - current;
			LP2LSP(a,w,q[current],fout);
			interpolate_LSP(q, current, q_sf);
			LSP2LP(q_sf[0], a_h[0]);
			LSP2LP(q_sf[1], a_h[1]);	
			
			// Ching-Kai's part
			perceptual_weighting(s, a_h[0], start, head, &r[0], s_w);
			perceptual_weighting(s, a_h[1], (start+SUBFRAMESIZE)%BUFFERSIZE, (head+SUBFRAMESIZE)%320, &r[40], s_w);
			open_loop_pitch_analysis(s_w, head, &T_op);
			response(a_h[0], &r[0], &h[0], &x[0]);
			response(a_h[1], &r[40], &h[40], &x[40]);
			head = (head + FRAMESIZE)%320;

			// Howard's part
			codebook_search(x, h, u, r, T_op, &P0, &P1, &P2, &g_p1, &g_p2, v, y, &S1, &S2, &C1, &C2, c, &g_c1, &g_c2, U_h);

			n   = (n+FRAMESIZE)%BUFFERSIZE;
			Outfile = fopen("output.out", "ab");
			fwrite(Out, sizeof(short), FRAMESIZE, Outfile);
			fclose(Outfile);
			In  = x_in + n;
			Out = y_out + n;
		}

		preprocessing(y_out, x_in, n, read);
		frame++;		
		for (i=0; i<read; i++)
			fprintf(fout[0],"%f\n",(float)x_in[(n+i)%BUFFERSIZE]);
		for (i=read; i<FRAMESIZE; i++)
			fprintf(fout[0],"%f\n",0);		

		start = (start+FRAMESIZE)%240;
		for (i=0; i<read; i++)
			s[(start+SUBFRAMESIZE+i)%240] = (float)y_out[(n+i)%BUFFERSIZE];
		for (i=read; i<FRAMESIZE; i++)
			s[(start+SUBFRAMESIZE+i)%240] = 0;
		window_and_autocorrelate(s,start,s_p,a,fout);
		current = 1 - current;
		LP2LSP(a,w,q[current],fout);
		interpolate_LSP(q, current, q_sf);
		LSP2LP(q_sf[0], a_h[0]);
		LSP2LP(q_sf[1], a_h[1]);

		// Ching-Kai's part
		perceptual_weighting(s, a_h[0], start, head, &r[0], s_w);
		perceptual_weighting(s, a_h[1], (start+SUBFRAMESIZE)%BUFFERSIZE, (head+SUBFRAMESIZE)%320, &r[40], s_w);
		open_loop_pitch_analysis(s_w, head, &T_op);
		response(a_h[0], &r[0], &h[0], &x[0]);
		response(a_h[1], &r[40], &h[40], &x[40]);
		head = (head + FRAMESIZE)%320;

		// Howard's part
		codebook_search(x, h, u, r, T_op, &P0, &P1, &P2, &g_p1, &g_p2, v, y, &S1, &S2, &C1, &C2, c, &g_c1, &g_c2, U_h);

		printf("number of frames = %d\n", frame);
		printf("The encoder input is the wave file %s\n",fname);
		printf("The encoder output are the files:\n");
		printf("  1) output.out  [The preprocessed signal]\n");
		printf("  2) LSF.dat     [The Line Spectral Frequencies of each frame]\n");
		printf("All other data files are created for the Matlab demo\n");
		n   = (n+read)%BUFFERSIZE;
		Outfile = fopen("output.out", "ab");
		fwrite(Out, sizeof(short), read, Outfile);
		fclose(Outfile);
		
		fclose(Infile);

		for (i=0; i<=6; i++)
			fclose(fout[i]);
	}

	fclose(fout2);
	
	return 0;
}