//
//	Decoder for G.729A CS-ACELP Speech Codec
// 
//		postprocessing.cpp : Module for 
//			- [A.4.2.1] Long-term postfilter
//			- [A.4.2.2] Short-term postfilter
//			- [A.4.2.3] Tilt Compensation
//			- [A.4.2.5] High-pass filtering and upscaling
//
#include "StdAfx.h"
#include "post_processing.h"

#include <limits.h>

static float r_p[6];
/************************************************************************
/*
/*	  Function to do all the post filtering [A.4.2]
/*    Input: t0		 Pitch delay
/*			 a_h[10] LPC coefficients
/*			 start	 start of the signal subframe
/*			 s[40]	 Reconstructed speech signal
/*    Output: sf[40] postfiltered reconstructed speech signal
/*            
/*
/************************************************************************/
void post(int T1, int T2, float a_h[2][10], float s[240], float s_ltp[90], float sf[90], float sf_p[90], short x[90], float *g, float r[240])
{
	int start;
	int n;

	start = 160;
	ltp_filter(T1, a_h[0], start, s, &s_ltp[0], r);	
	start = 10;
	stp_filter(a_h[0], start, s_ltp, sf);
	adaptive_gain_control(&s[160], &sf[10], &sf_p[10], g);

	start = 200;
	ltp_filter(T2, a_h[1], start, s, &s_ltp[40], r);	
	start = 50;
	stp_filter(a_h[1], start, s_ltp, sf);
	adaptive_gain_control(&s[200], &sf[50], &sf_p[50], g);

	for (n=0; n<10; n++)
	{
		s_ltp[n] = s_ltp[n+80];
		sf[n] = sf[n+80];		
	}

	start = 10;
	highpass_upscaling(sf_p, x, start);

	for (n=0; n<10; n++)
	{
		sf_p[n] = sf_p[n+80];
		x[n] = x[n+80];
	}
	
	for (n=0; n<80; n++)
	{
		x[n+10] *= 2;
	}

	//update_s(start, s, s_n);
	//tilt_filter(a_h, start, s, s_n);
}

/************************************************************************
/*
/*	  Function to do the long term post filtering [A.4.2.1]
/*    Input: t0		 Pitch delay
/*			 a_h[10] LPC coefficients
/*			 start	 start of the signal subframe
/*			 s[240]	 Reconstructed speech signal
/*    Output: s_n[240]	 long-term postfiltered speech signal
/*            
/*
/************************************************************************/
void ltp_filter(int t0, float a_h[10], int start, float s[240], float s_n[50], float r[240])
{
	float a_p[10];
	float gamma = GAMMA_N;
	int n, k;
	int max_k;
	float max_Rk;
	int pos, posD;

	//Compute the A'(z) = A^(z/gamma)
	//a_p[n] = (gamma)^n * a_h[n]
	for (n=0; n<10; n++)
	{
		a_p[n] = gamma*a_h[n];
		gamma *= GAMMA_N;
	}

	//Compute the residual signal r(n) for n = 0...39
	for (n=0; n<40; n++)
	{
		float sum = 0;
		for (int i=0; i<10; i++)
		{
			pos = (start+n-i-1)%240;
			sum += a_p[i]*s[pos];
		}
		pos = (start+n)%240;
		r[pos] = s[pos] + sum;
	}
	
	//Find the best integer T
//	max_k = t0-3;
//	max_Rk = (float)INT_MAX;
	for (k=t0-3; k<=t0+3; k++)
	{
		float Rk = 0;
		for (n=0; n<40; n++)
		{
			pos = (start+n)%240;
			posD = (start+n-k+240)%240;
/*			if (n-k <0)
				Rk +=(r[n]*r_p[abs(n-k)-1]);
			else
*/
			Rk += (r[n]*r[n-posD]);
		}
		if (k==t0-3)
		{
			max_k = k;
			max_Rk = Rk;
		}
		else if (max_Rk < Rk)
		{
			max_k = k;
			max_Rk = Rk;
		}
	}
/*
	for (k=0; k<6; k++)
		r_p[k] = r[40-k-1];
*/
	//Problem??? gl not used now
	//Don't know where to get the gl value
	//Long-Term postfilter Hp(z) = (1+GAMMA_P*Z^-T)/(1+GAMMA_P)
	for (n=0; n<40; n++)
	{
		pos = (start+n)%240;
		posD = (start+n-max_k+240)%240;
		s_n[n+10] = (s[pos]+GAMMA_P*s[posD])/(1+GAMMA_D);
	}
}

/************************************************************************
/*
/*	  Function to do the short term post filtering  [A.4.2.2]
/*		   and to do the tilt compensation			[A.4.2.3]
/*    Input: a_h[10] LPC coefficients
/*			 start	 start of the signal subframe
/*			 s[240]	 Reconstructed speech signal
/*    Output: s_n[240]	 long-term postfiltered speech signal
/*            
/*
/************************************************************************/
void stp_filter(float a_h[10], int start, float s[90], float s_n[90])
{
	float a_d[10];
	float a_n[10];
	float hf[90];
	float delta[90];
	float gamma = GAMMA_N;
	int pos, posD;
	int n;


	for (n=0; n<90; n++)
	{
		hf[n] = 0;
		delta[n] = 0;
	}
	delta[start] = 1;
	
	//a_n[n] = (gamma)^n * a_h[n]
	for (n=0; n<10; n++)
	{
		a_n[n] = gamma*a_h[n];
		gamma *= GAMMA_N;
	}
	gamma = GAMMA_D;
	//a_d[n] = (gamma)^n * a_h[n]
	for (n=0; n<10; n++)
	{
		a_d[n] = gamma*a_h[n];
		gamma *= GAMMA_D;
	}
	
	//Problem??? Input and Output for initial value
	//Filtering of Hf(z) = A^(z/GAMMA_N)/A^(z/GAMMA_D)
	hf[start] = 1;
	for (n=0; n<40; n++)
	{
		int i;
		pos = start+n;
		float sumx = 0;
		hf[pos] = delta[pos];
		for (i=0; i<10; i++)
		{
			posD = start+n-i-1;
			sumx += a_n[i]*s[posD];
			hf[pos] += a_n[i]*delta[posD];
		}
		float sumy = 0;
		float sumhf = 0;
		for (i=0; i<10; i++)
		{
			posD = start+n-i-1;
			sumy += a_d[i]*s_n[posD];			
			sumhf += a_d[i]*hf[posD];
		}		
		s_n[pos] = (s[pos] + sumx) - sumy;
		hf[pos] -= sumhf;
	}

//	update_s(start, s, s_n);
	//Calculate rh(1) and rh(0)
	float rh1 = 0;
	float rh0 = 0;
	for (int j=0; j<=20; j++)
	{
		rh1 += hf[start+j]*hf[start+j+1];
		rh0 += hf[start+j]*hf[start+j];
	}
	rh0 += hf[start+21]*hf[start+21];

	//k' = -rh(1)/rh(0)
	float k1 = -rh1/rh0;
	//if k'<0 then  Ht(z) = 1+GAMMA_T*k'z^-1
	for (n=0; n<40; n++)
	{
		pos = start+n;
		posD = start+n-1;
		if (k1<0)
			s_n[pos] = s[pos]+ GAMMA_T*k1*s[posD];
		else 
			s_n[pos] = s[pos];
	}
}

void highpass_upscaling(float y[90], short x[90], int start)
{
	int i;
	int n0, n1, n2;

	for (i=0; i<80; i++)
	{
		n0 = i+start;
		n1 = i-1+start;
		n2 = i-2+start;
		x[n0] = (short)(B0*(double)y[n0] + B1*(double)y[n1] + B2*(double)y[n2]+
						A1*(double)x[n1] + A2*(double)x[n2]);
	}
}

void adaptive_gain_control(float s_h[40], float sf[40], float sf_p[40], float *g)
{
	float G;
	float temp;
	int n;

	for (n=0, G=0; n<40; n++)
		G += s_h[n] * s_h[n];

	for (n=0, temp=0; n<40; n++)
		temp += sf[n] * sf[n];

	G = sqrt(G/temp);

	for (n=0; n<40; n++)
	{
		*g = 0.9* *g + 0.1*G;
		sf_p[n] = *g * sf[n]; 
	}
}

